/**********************************************************************
 *                        INPUT/OUTPUT FUNCTION                       *
 **********************************************************************/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <limits.h>
#include "patricia.h"

/******************************************
 *           GLOBAL VARIABLEs             *
 ******************************************/
 
char  FNAME[256]="";
char  ONAME[256]="";
FILE* out;
FILE* fptrans_read;
long  max_item=MAX_ELEM;
int   max_trans_len=0;
int*  count_element;
int   min_sup; 

/******************************************
 *               pat_error                *
 ******************************************/

void pat_error(char * s) {

  printf("Error: %s\n", s);
  fflush(stdout);
  exit(1);
}

/******************************************
 *               init_count()             *
 ******************************************/

void init_count() {
  if((count_element=(int *) calloc(MAX_ELEM,sizeof(int)))==NULL)
    pat_error("Error: count_element array allocation");
}

/******************************************
 *              read_param                *
 ******************************************/

void read_param(int argc, char *argv[]) {

  char *pn = NULL; 
  pn = argv[0]; /*Program name*/  
  if (argc < 3 || argc > 4){
   usage(pn);
   exit(1);
  }       
  /* Set FNAME as input file name */
  strcpy(FNAME,"");
  strcat(FNAME, argv[1]);
  /* Set min_sup  */	    
  min_sup = atoi(argv[2]);	
 /* Set FNAME as output file name */
  if(argc==4){
   strcpy(ONAME,"");
   strcat(ONAME, argv[3]);
  }
  /* Verify only FNAME and min_sup. ONAME is optional */
  if(!( (min_sup >0 ) && (strcmp(FNAME,"")!=0))){
	 usage(pn);
	 exit(1);
  }
  min_sup--;
	
}

/******************************************
 *                 usage                  *
 ******************************************/

void usage(char *prog_name) {

  printf("# usage: %s datafile minsup [output]\n", prog_name);
  printf("#    datafile : name of the file of transactions\n"); 
  printf("#    minsup   : count      support (> 0)\n");
  printf("#    [output] : name of the output file for frequent itemset [optional]\n");
}


/******************************************
 *               printSet                 *
 ******************************************/

void printSet(int length, int *iset, int support){

  int i;

  /* Print itemset only if output file is open */
   for(i=0; i<length; i++)
    fprintf(out, "%d ", iset[i]);
    fprintf(out, "(%d)\n", support);
  
}

/******************************************
 *               read_int                 *
 ******************************************/

int read_int(int *item, char *c){
   
  (*c)=getc(fptrans_read);
  if (((*c)>='0') && ((*c) <='9')){
    (*item)*=10;
    (*item) += (int)(*c)-(int)('0');
    (*c)=getc(fptrans_read); 
  }
  else return 0;
  while(((*c)>='0') && ((*c) <='9')){
    (*item)*=10;
    (*item) += (int)(*c)-(int)('0');
    (*c)=getc(fptrans_read);
  }
  return 1;
}

/******************************************
 *               init_out                 *
 ******************************************/

void init_out() {

  /*Open the output file only if ONAME != "" */
  if(strcmp(ONAME,"")!=0){
    if ((out = fopen(ONAME, "w")) == NULL) {
      fprintf (stderr, "init_out. Error opening %s: %s.\n", "frequent_itemset", strerror(errno));
      exit (-1);
    }
  }
}

/******************************************
 *          init_transaction_read         *
 ******************************************/

void init_transaction_read() {

  if ((fptrans_read = fopen(FNAME, "r")) == NULL) {
    fprintf (stderr, "init_transaction_read. Error opening %s: %s.\n", FNAME, strerror(errno));
    exit (-1);
  }
}

/******************************************
 *          done_transaction_read         *
 ******************************************/

void done_transaction_read() {
  fclose (fptrans_read);
}

/******************************************
 *                done_write              *
 ******************************************/

void done_write() {
  if(out) fclose (out); /* Close out only if open */
}

/******************************************
 *          getNextTransaction            *
 ******************************************/
 
int getNextTransaction(TRANSACTION *t){

  char c;
  int d;
  int pos=0,len=0,i;

  /* First dataset scan ==> t==NULL */
  if (t == NULL) {
    do {
      /* For each item in transaction, update count_element and max_item */ 
      int item=0;
      if(read_int(&item,&c)) {
        if (item > max_item){
          /* Expand count_element vector */
          max_item=item*2; 
          if((count_element=(int *)realloc(count_element,max_item))==NULL)
	   pat_error("Error: count_element reallocation in first read of getNextTransaction");
        }
        count_element[item]++;
        len++; /* transaction length */
      }
    } while(c!= '\n' && !feof(fptrans_read));

    /* If end of file reached: rewind */
    if(feof(fptrans_read)){
       rewind(fptrans_read);
       return 0;
    }

    /* Update max_trans_len */
    if(len > max_trans_len) max_trans_len=len;
    return 1;
  }
  else { /* t<>NULL ==> Second scan of dataset */
    i=1;
    do {
      /* 
       * Read transaction from dataset into t, limited
       * to frequent items, and representing items in t 
       * through their indices in IL
       */
      int item=0;
      if(read_int(&item,&c)) {
	if(count_element[item]>=0){ 
          (*t)[i++]=count_element[item];
        }
      }
    } while(c!= '\n' && !feof(fptrans_read));

    /* Set transaction's size and multiplicity */
    (*t)[0]=i-1;
    (*t)[(*t)[0]+1]=1;

    /* Sort transaction in ascending order */
    Tran_sort(*t, 1, i-1);

    if(feof(fptrans_read)) {
      done_transaction_read();
      return 0; /* end of input file reached */ 
    }
    return 1;
  } /* end if (t==NULL) */
}
